/*****************************************************************************\
* (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// Include files

// from Gaudi
#include "GaudiKernel/IRegistry.h"

#include "Kernel/IEventTupleTool.h" // Interface

// local
#include "EventTuple.h"

//-----------------------------------------------------------------------------
// Implementation file for class : EventTuple
//
// 2008-07-01 : Patrick Koppenburg
//-----------------------------------------------------------------------------

// Declaration of the Algorithm Factory
DECLARE_COMPONENT( EventTuple )

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
EventTuple::EventTuple( const std::string& name, ISvcLocator* pSvcLocator ) : GaudiTupleAlg( name, pSvcLocator ) {
  m_toolList.push_back( "TupleToolEventInfo" );
  m_collectionName = name + "/ETC";
  declareProperty( "ToolList", m_toolList, "List of tools to be used" );
  declareProperty( "TupleName", m_tupleName = "EventTuple", "Name of nTuple" );
  declareProperty( "CollectionName", m_collectionName = "", "Default is evtColPath()" );
}
//=============================================================================
// Destructor
//=============================================================================
EventTuple::~EventTuple() {}

//=============================================================================
// Initialization
//=============================================================================
StatusCode EventTuple::initialize() {
  const StatusCode sc = GaudiTupleAlg::initialize();
  if ( sc.isFailure() ) return sc;
  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Initialize" << endmsg;

  info() << "Tools to be used : ";
  for ( std::vector<std::string>::const_iterator s = m_toolList.begin(); s != m_toolList.end(); ++s ) {
    m_tools.push_back( tool<IEventTupleTool>( *s, this ) );
    info() << *s << ", ";
  }
  info() << endmsg;

  if ( produceEvtCols() ) {
    if ( m_collectionName.empty() ) { m_collectionName = evtColPath(); }
    info() << "Will be writing an ETC with name " << m_collectionName << "/" << m_tupleName << endmsg;
  }

  return sc;
}

//=============================================================================
// Main execution
//=============================================================================
StatusCode EventTuple::execute() {

  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Execute" << endmsg;
  ++counter( "Event" );
  StatusCode sc = StatusCode::SUCCESS;

  Tuple tuple = ( produceEvtCols() ? evtCol( m_tupleName, m_collectionName ) : nTuple( m_tupleName ) );
  if ( msgLevel( MSG::VERBOSE ) ) verbose() << "Got tuple" << endmsg;
  if ( produceEvtCols() ) {
    // pick up the location of the event --
    // this is what makes the tag collection a collection...
    DataObject* pObject = get<DataObject>( "/Event" );
    if ( pObject ) {
      if ( msgLevel( MSG::VERBOSE ) ) verbose() << "Filling Address " << endmsg;
      sc = tuple->column( "Address", pObject->registry()->address() );
      if ( !sc ) {
        err() << "Error writing address" << endmsg;
        return sc;
      }
    } else {
      return Error( "    not able to retrieve IOpaqueAddress" );
    }
  }

  sc = tuple->column( "EventInSequence", (unsigned long long)counter( "Event" ).nEntries() );
  if ( sc.isFailure() ) return sc;

  for ( std::vector<IEventTupleTool*>::iterator i = m_tools.begin(); i != m_tools.end(); ++i ) {
    if ( msgLevel( MSG::VERBOSE ) ) verbose() << "Filling " << ( *i )->name() << endmsg;
    sc = ( *i )->fill( tuple );
    if ( msgLevel( MSG::VERBOSE ) ) verbose() << ( *i )->name() << " returns " << sc << endmsg;
    if ( !sc ) return sc;
  }
  if ( msgLevel( MSG::VERBOSE ) ) verbose() << "Writing tuple" << endmsg;
  return tuple->write();
}

//=============================================================================
