/*****************************************************************************\
* (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// $Id: $
#ifndef MCDISPLVERTICES_H
#define MCDISPLVERTICES_H 1

// Include files
// from DaVinci, this is a specialized GaudiAlgorithm
#include "Kernel/DaVinciTupleAlgorithm.h"

// To associate a particle to the true one.
#include "Kernel/Particle2MCLinker.h"

// To recreate track measurements
#include "TrackInterfaces/IMeasurementProvider.h"

// To use Wouter's vertex tool
#include "TrackInterfaces/ITrackVertexer.h"

// To decode L0DU raw infos
#include "L0Interfaces/IL0DUFromRawTool.h"

// Hlt tools
#include "Kernel/ITriggerTisTos.h"

// Detector description
#include "DetDesc/IDetectorElement.h"
#include "DetDesc/IGeometryInfo.h"
#include "DetDesc/ILVolume.h"
#include "DetDesc/IPVolume.h"
#include "DetDesc/ITransportSvc.h"
#include "DetDesc/Material.h"

/** @class MCDisplVertices MCDisplVertices.h
 *
 *  @brief Apply a set of cuts on given Particles.
 *  @author Neal Gauvin
 *  @date  20 september 2010
 */

class MCDisplVertices : public DaVinciTupleAlgorithm {

public:
  /// Standard constructor
  MCDisplVertices( const std::string& name, ISvcLocator* pSvcLocator );

  virtual ~MCDisplVertices(); ///< Destructor

  StatusCode initialize() override; ///< Algorithm initialization
  StatusCode execute() override;    ///< Algorithm execution
  StatusCode finalize() override;   ///< Algorithm finalization

private:
  // A list is a sequence container that support rapide insertion or deletion
  // at any position but does not support random access.
  typedef std::vector<LHCb::Particle> VecPart;
  typedef std::vector<LHCb::Vertex>   VecVert;

  Particle2MCLinker*    m_pLinker;
  const IVertexFit*     m_vFit;
  IMeasurementProvider* m_measProvider;
  ITrackVertexer*       m_tFit; // Wouter's tool
  ITriggerTisTos*       m_tisTos;
  IL0DUFromRawTool*     m_l0BankDecoder;
  ITransportSvc*        m_transSvc;
  IGeometryInfo*        m_lhcbGeo;

  // Studies
  void StudyActivity();
  void StudyDistanceBet2Interactions();
  void StudyDiffGenMC();
  void StudyPV();
  void StudyHltEfficiency( const Gaudi::XYZPoint& );
  void StudyPreyIP( const LHCb::Particle*, const LHCb::RecVertex* );
  void StudyDispersion( const LHCb::Particle* );
  void StudyDispersion( const LHCb::Particle*, const LHCb::Particle* );
  void StudyDgtsIP( const LHCb::Particle* );
  void StudyDgtsIP( const LHCb::RecVertex* );
  void StudyEoverNbTrk();
  void StudyPreyComposition( const LHCb::Particle*, std::string );
  void StudyParticleGun( LHCb::Particle::ConstVector& );

  // HepMC and MC functions
  StatusCode  GetHepMCInfos();
  void        GetHepMCStable( HepMC::GenParticle*, int&, HepMC::ThreeVector* );
  void        GetHepMCStable( HepMC::GenParticle*, int&, HepMC::GenVertex* );
  std::string Print( HepMC::ThreeVector );
  std::string Print( HepMC::FourVector );

  StatusCode GetMCInfos();
  bool       GetMCPrey( const LHCb::Particle* );
  void       GetMCStable( const LHCb::MCVertex*, std::string );
  void       GetMCStable( const LHCb::MCVertex* );
  void       GetMCStable( const LHCb::MCParticle*, const Gaudi::XYZPoint&, LHCb::MCParticle::ConstVector& );
  void       GetMCStable( const LHCb::MCParticle*, const Gaudi::XYZPoint&, unsigned int& );
  void       GetMCStable( const LHCb::MCParticle*, unsigned int& );
  void       GetMCStable( const LHCb::MCVertex*, int&, Gaudi::LorentzVector& );
  void       SaveGenPartinTuple( const HepMC::GenEvent* );
  void       SaveMCPartinTuple( LHCb::MCParticles* );
  StatusCode SaveGEC( Tuple&, LHCb::Particle::ConstVector& );
  StatusCode SavePVs( Tuple& );

  // questions
  double WhichMCPrey( const LHCb::Particle* );
  double IsitFromaPrey( const LHCb::MCParticle*, const LHCb::MCVertex* );
  bool   IsitFromaPrey( const LHCb::MCParticle* );
  bool   IsitFromaPrey( HepMC::GenParticle* );
  bool   IsNeutrino( const LHCb::MCParticle* );
  bool   IsNeutrino( HepMC::GenParticle* );
  bool   IsQuark( HepMC::GenParticle* );
  bool   IsAPointInDet( const LHCb::Particle*, int mode = 2, double range = 1 * Gaudi::Units::mm );
  bool   IsAPointInDet( const Gaudi::XYZPoint& );
  /// Is a point in the RF-Foil ?
  bool IsInRFFoil( const Gaudi::XYZPoint& );
  /// Has the (MC) vertex been generated by collisin of flying particles with air of beam pipe ?
  bool IsFromAir( const LHCb::Particle* );
  /// Has the given Particle a related MCVertex from int with matter ?
  const LHCb::MCVertex* HasIntInAncestors( const LHCb::Particle* );
  /// Has a RecVertex a backward track ?
  bool HasBackwardTracks( const LHCb::Particle* );
  bool HasBackwardTracks( const LHCb::RecVertex* );
  /// Has a RecVertex a backward and a forward track ?
  bool HasBackAndForwardTracks( const LHCb::RecVertex* );
  /// Has a candidate a daughter muon ?
  double                  HasMuons( const LHCb::Particle* );
  const LHCb::MCParticle* WhichMother( const LHCb::Particle* );
  const LHCb::MCParticle* WhichMother( const LHCb::MCParticle* );

  // Geometric tools
  double          Mult( const Gaudi::XYZPoint&, const Gaudi::XYZPoint& );
  double          Mult( const Gaudi::LorentzVector&, const Gaudi::LorentzVector& );
  Gaudi::XYZPoint Plus( const Gaudi::XYZPoint&, const Gaudi::XYZPoint& );
  Gaudi::XYZPoint Minus( const Gaudi::XYZPoint&, const Gaudi::XYZPoint& );
  double          Norm( const Gaudi::LorentzVector& );
  double          Norm( const Gaudi::XYZPoint& );
  double          Norm( const Gaudi::XYZVector& );
  Gaudi::XYZPoint Normed( const Gaudi::LorentzVector&, double range = 1 );
  double          VertDistance( const Gaudi::XYZPoint&, const Gaudi::XYZPoint& );
  double          VertDistance( HepMC::ThreeVector&, HepMC::ThreeVector& );
  double          VertDistance( HepMC::GenVertex*, HepMC::GenVertex* );
  double          VertDistance( HepMC::ThreeVector*, LHCb::Vertex* );
  double          VertDistance( HepMC::ThreeVector*, HepMC::GenVertex* );
  double          VertDistance( HepMC::ThreeVector*, const LHCb::RecVertex* ) const;
  double          VertDistance( HepMC::ThreeVector*, const LHCb::RecVertex& ) const;
  double          VertDistance( HepMC::ThreeVector&, LHCb::RecVertex& );
  double          VertDistance( HepMC::ThreeVector*, HepMC::ThreeVector& );
  double          VertDistance( HepMC::ThreeVector& ); // return distance to ref frame.
  double          Angle( const Gaudi::LorentzVector&, const Gaudi::LorentzVector& );
  double          Angle( const Gaudi::XYZVector&, const Gaudi::XYZVector& );
  double          ImpactParam( const LHCb::MCParticle*, const LHCb::MCVertex* );
  double          ImpactParam( const Gaudi::XYZPoint&, const Gaudi::LorentzVector&, const Gaudi::XYZPoint& );
  double          ImpactParam( const Gaudi::XYZPoint&, const Gaudi::XYZVector&, const Gaudi::XYZPoint& );
  void            Resolution( const Gaudi::XYZPoint& );
  void            Resolution();

  unsigned int GetNbVeloTracks();
  double       GetSumPt( const LHCb::Particle* );
  void         GetQPt( const LHCb::Vertex* v, double& sumpt, double& sumq, bool weight = false );
  double       GetCharge( const LHCb::Vertex* v, bool weight = false );
  double       GetRecMass( const LHCb::Particle::ConstVector& );
  double       GetRecMass( const LHCb::Particle::ConstVector&, double& );
  double       GetPt( const LHCb::RecVertex* RV );
  double       GetPt( const Gaudi::LorentzVector&, const Gaudi::LorentzVector& );
  double       GetDeltaPhi( const Gaudi::XYZPoint&, const Gaudi::XYZPoint& );
  double       GetDeltaR( const LHCb::Particle*, const LHCb::Particle* );
  double       GetR( const Gaudi::XYZPoint& );
  double       GetMCMass( LHCb::MCParticle::ConstVector& );
  double       GetRFromBL( const Gaudi::XYZPoint& );
  void         GetMassFromLongTracks( const LHCb::Particle*, double&, int& );
  void         GetUpstreamPV(); ///< Get the Upstream PV
  void         GetPVs();        ///< Get the PV candidates
  void GetPartsFromRecVtx( const LHCb::RecVertex*, const LHCb::Particle::ConstVector&, LHCb::Particle::ConstVector& );
  StatusCode SaveCaloInfos( Tuple& );
  StatusCode GetCaloInfos( std::string, double&, double& );
  StatusCode GetCaloClusterInfos( std::string, Tuple&, double&, double& );
  StatusCode SaveTrigInfinTuple( Tuple& );
  StatusCode SaveL0RawInfos( Tuple& );
  StatusCode fillHeader( Tuple& );
  StatusCode BackToBack( LHCb::Particle::ConstVector& );

  /***************************************************************/ /**
                                                                     * the type of radial (R) cut to be applied
                                                                     * ""                   : cut with respect to
                                                                     *(0,0,z) "FromUpstreamPV"     : cut with respect to
                                                                     *the upstream PV (PV3D) "FromBeamLine"       : cut
                                                                     *with respect to given beam line
                                                                     ******************************************************************/
  std::string     m_RCut;
  std::string     m_BLLoc; ///< Location in TES of Beam line
  LHCb::Particle* m_BeamLine;
  bool            m_SaveTuple;                                      ///< Save prey properties in a tuple
  bool            m_SaveTrigInfos;                                  ///< Save trigger infos in a tuple
  bool            m_SaveonTES;                                      ///< Save Event on TES if fires
  /***************************************************************/ /**
                                                                     * World volume simulated with Air in MC Data
                                                                     * Therefore vacuum in beam pipe simulated with air
                                                                     *at p atm. Set m_RemVtxFromAir=True to remove vtx
                                                                     *generated by collision of flying particles with
                                                                     *air of beam pipe.
                                                                     ******************************************************************/
  bool         m_RemVtxFromAir;
  const double pi;
  /// Proportion of preys in different purity areas
  std::vector<double> m_purities;
  double              m_maxunpure;
  double              m_minmixed;
  double              m_maxmixed;
  // cuts
  double       m_PreyMinMass;                                       ///< Minimum reconstructed mass
  double       m_PreyMaxMass;                                       ///< Maximum reconstructed mass
  double       m_SumPt;                                             ///< Sumpt of all daughters tracks
  double       m_DistMax;                                           ///< Max distance of the vertices
  double       m_RMin;                                              ///< Min dist to the z axis
  double       m_RMax;                                              ///< Max dist to the z axis
  double       m_MaxChi2OvNDoF;                                     ///< Max chi2 of a vertex
  double       m_PurityMin;                                         ///< Min purity requested, MC set on
  double       m_MuonpT;                                            ///< Muon with a pT of min m_MuonpT
  double       m_DocaMax;                                           ///< Max distance of closest approach
  unsigned int m_NbCands;                                           ///< Min nb of desired candidates
  unsigned int m_PVnbtrks;                                          ///< Min nb of trks for upstream PV candidate
  int          m_nTracks;                                           ///< Min # of tracks at reconstructed vertex
  double       m_SigmaZ;                                            ///< Max on error Z
  double       m_SigmaR;                                            ///< Max on error R
  double       m_MinX;                                              ///< Min X position
  double       m_MaxX;                                              ///< Max X position
  double       m_MinY;                                              ///< Min Y position
  double       m_MaxY;                                              ///< Max Y position
  double       m_MinZ;                                              ///< Min Z position
  double       m_MaxZ;                                              ///< Max Z position
  /***************************************************************/ /**
                                                                     * Remove vtx if in detector material ?
                                                                     * if = 0  : disabled
                                                                     * if = 1  : remove reco vtx if in detector material
                                                                     * if = 2  : remove reco vtx if rad length from
                                                                     *decay pos - DetDist to decay pos + DetDist along
                                                                     *momentum is > threshold if = 3 : remove reco vtx
                                                                     *if rad length along
                                                                     *                             +- DetDist *
                                                                     *PositionCovMatrix if = 4 : 3 but range+3 if in RF
                                                                     *foil.
                                                                     ******************************************************************/
  int    m_RemVtxFromDet;
  double m_DetDist; ///< Min distance to det material
  ///< Remove vtx if found in RF-Foil area, based on geometric cuts
  bool   m_RemFromRFFoil;
  double m_Backtoback;

  Gaudi::Transform3D m_toVeloLFrame; ///< to transform to local velo L frame
  Gaudi::Transform3D m_toVeloRFrame; ///< to transform to local velo R frame

  bool                                m_MC;      ///< Retrieve MC infos or not
  std::string                         m_HepMC;   ///< Retrieve HepMC infos or not
  std::vector<Gaudi::XYZPoint>        m_MCSVPos; // on array of ptr on MCSV
  std::vector<Gaudi::XYZPoint>        m_MCPos;   // on array of ptr on MCPrey
  std::vector<const LHCb::RecVertex*> PVs;       ///< The PV candidates
  const LHCb::RecVertex*              PV;        // The Primary Vertex !
  const LHCb::MCVertex*               MCPV;      // The MC Primary Vertex !

  std::string      m_Prey;             ///< LHCb Name of the prey
  LHCb::ParticleID m_PreyID;           // PDG ID of this particle
  std::string      m_MotherPrey;       ///< LHCb Name of the mother of the prey
  LHCb::ParticleID m_MotherPreyID;     // PDG ID of this particle
  bool             m_IsPreyFromMother; ///< Is the Prey from the given Mother ?

  // Dummy things for HepMC
  HepMC::FourVector dumV1; // dummy lorentz vector
  HepMC::FourVector dumV2; // dummy lorentz vector

  struct sortPVdz {
    double refz;
    sortPVdz() { refz = 0.; };
    bool operator()( const LHCb::RecVertex* first, const LHCb::RecVertex* second ) {

      return std::abs( first->position().z() - refz ) < std::abs( second->position().z() - refz );
    }
  } SortPVdz;
  /// To sort PVs with ascending z position
  struct sortPVz {
    bool operator()( const LHCb::RecVertex* first, const LHCb::RecVertex* second ) {
      return first->position().z() < second->position().z();
    }
  } SortPVz;
};

#endif // MCDISPLVERTICES_H
