/*****************************************************************************\
* (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
#ifndef LOKI_MCMATCHER_H
#define LOKI_MCMATCHER_H 1
// ============================================================================
// Include files
// ============================================================================
// LoKiMC
// ============================================================================
#include "LoKi/IMCDecay.h"
// ============================================================================
// LoKiPhysMC
// ============================================================================
#include "LoKi/MCTruth.h"
// ============================================================================
// forward declaration
// ============================================================================
class GaudiAlgorithm;
// ============================================================================
namespace LoKi {
  // ==========================================================================
  /// forward declarations
  class MCMatch; // forward declarations
  // ==========================================================================
  namespace PhysMCParticles {
    // ========================================================================
    enum ProtoPMatch {
      All,         // Neutral , Charged + Upstream
      Charged,     //           Charged + Upstream
      ChargedLong, //           Charged
      Neutral
    };
    // ========================================================================
    /** @class MCMatcherBase LoKi/MCMatcher.h
     *
     *  The special version of class LoKi::PhysMCParticles::MCTruth
     *  to be used withing CombineParticles/DecayTree frameworks
     *
     *  @author Vanya  BELYAEV Ivan.Belyaev@nikhef.nl
     *  @date   2009-08-11
     */
    class MCMatcherBase : public LoKi::PhysMCParticles::MCTruth {
    public:
      // ======================================================================
      /// the actual type for TES-locations for relation tables
      typedef std::vector<std::string> Locations; // TES-locations of tables
      // ======================================================================
      typedef Decays::iNode iNode;
      // ======================================================================
    public:
      // ======================================================================
      /// constructor from the list of TES-locations
      MCMatcherBase( Locations locations, std::string thehead = "" );
      /// constructor from the list of TES-location
      MCMatcherBase( const std::string& location ) : MCMatcherBase( Locations{location} ) {}
      /// constructor from the list of TES-locations
      MCMatcherBase( const std::string& location1, const std::string& location2 )
          : MCMatcherBase( Locations{location1, location2} ) {}
      /// constructor from the list of TES-locations
      MCMatcherBase( const std::string& location1, const std::string& location2, const std::string& location3 )
          : MCMatcherBase( Locations{location1, location2, location3} ) {}
      /// constructor from protopartcile  matching flag
      MCMatcherBase( const ProtoPMatch& protoMatch, std::string thehead = "" );
      /// copy constructor
      MCMatcherBase( const MCMatcherBase& right );
      // ======================================================================
      /// virtual destructor
      virtual ~MCMatcherBase();
      // ======================================================================
    protected:
      // ======================================================================
      /// load the data
      StatusCode load() const; // load the data
      /// get MC-particles
      virtual StatusCode getMCParticles( const LHCb::MCParticle::Container& cnt ) const = 0;
      /// clear all storages
      void clearAll() const;
      /// get all locations
      const Locations& locations() const { return m_locations; }
      /// get the MC head
      const std::string& head() const { return m_head; }
      /// get the algorithm
      const GaudiAlgorithm* alg() const { return m_alg; }
      /// set the algorithm
      void setAlg( const GaudiAlgorithm* a ) const { m_alg = a; }
      // ======================================================================
    private:
      // ======================================================================
      /// the default constructor is disabled
      MCMatcherBase(); // the default constructor is disabled
      /// the assignement operator is disabled
      MCMatcherBase& operator=( const MCMatcherBase& );
      // ======================================================================
    private:
      // ======================================================================
      /// TES locations of relation tables
      Locations m_locations; // TES locations of relation tables
      /// the location of MC-particles in TES
      std::string m_head; // the location of MC-particles in TES
      /// the algorithm
      mutable const GaudiAlgorithm* m_alg = nullptr; // the algorithm
      // ======================================================================
    };
    // ========================================================================
    /** @class MCSelMatch
     */
    class MCSelMatch : public MCMatcherBase {
    public:
      // ======================================================================
      /// the actual type of MC-selector
      typedef LoKi::Types::MCCuts MCCuts;
      // ======================================================================
    public:
      // ======================================================================
      /// Default copy constructor
      MCSelMatch( const MCSelMatch& ) = default;
      /** constructor from MC-selector and TES-locations
       *  @param selector Monte Carlo selector
       *  @param locations TES-locations of relation tables
       */
      MCSelMatch( const MCCuts& cuts, const Locations& locations, const std::string& thehead = "" );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param location TES-location of relation tables
       */
      MCSelMatch( const MCCuts& cuts, const std::string& location );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param location1 TES-location of relation tables
       *  @param location2 TES-location of relation tables
       */
      MCSelMatch( const MCCuts& cuts, const std::string& location1, const std::string& location2 );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param location1 TES-location of relation tables
       *  @param location2 TES-location of relation tables
       *  @param location3 TES-location of relation tables
       */
      MCSelMatch( const MCCuts& cuts, const std::string& location1, const std::string& location2,
                  const std::string& location3 );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param protoMatch the key for ProtoParticles match
       */
      MCSelMatch( const MCCuts& cuts, const ProtoPMatch& protoMatch = All, const std::string& thehead = "" );
      // ======================================================================
      /// MANDATORY: virtual destrcutor
      virtual ~MCSelMatch();
      /// MANDATORY: clone method ("virtual constructor")
      MCSelMatch* clone() const override;
      /// MANDATORY: the only one essential method
      result_type operator()( argument p ) const override;
      /// OPTIONAL: nice printout
      std::ostream& fillStream( std::ostream& ) const override;
      // ======================================================================
    protected:
      // ======================================================================
      /// get MC-particles
      StatusCode getMCParticles( const LHCb::MCParticle::Container& cnt ) const override;
      // ======================================================================
    private:
      // ======================================================================
      /// default constructor is disabled
      MCSelMatch(); // default constructor is disabled
      /// the assignement operator is disabled
      MCSelMatch& operator=( const MCSelMatch& );
      // ======================================================================
    private:
      // ======================================================================
      /// the actual selector of Monte Carlo particles
      LoKi::Types::MCCut m_cuts; // the selector of Monte Carlo particles
      // ======================================================================
    };
    // ========================================================================
    class MCTreeMatch : public MCMatcherBase {
    public:
      // ======================================================================
      // the actual type of decay finder
      typedef Decays::IMCDecay::Finder Finder;
      // the actual type of decay descriptor
      typedef Finder::iTree iTree;
      // ======================================================================
    public:
      // ======================================================================
      /// Default copy constructor
      MCTreeMatch( const MCTreeMatch& ) = default;
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param locations TES-locations of relation tables
       */
      MCTreeMatch( const iTree& decay, const Locations& locations, const std::string& thehead = "" );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param location TES-location of relation tables
       */
      MCTreeMatch( const iTree& decay, const std::string& location );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param location1 TES-location of relation tables
       *  @param location2 TES-location of relation tables
       */
      MCTreeMatch( const iTree& decay, const std::string& location1, const std::string& location2 );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param location1 TES-location of relation tables
       *  @param location2 TES-location of relation tables
       *  @param location3 TES-location of relation tables
       */
      MCTreeMatch( const iTree& decay, const std::string& location1, const std::string& location2,
                   const std::string& location3 );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param protoMatch the key for ProtoParticles match
       */
      MCTreeMatch( const iTree& decay, const ProtoPMatch& protoMatch = All, const std::string& thehead = "" );
      // ======================================================================
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param locations TES-locations of relation tables
       */
      MCTreeMatch( const std::string& decay, const Locations& locations, const std::string& thehead = "" );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param location TES-location of relation tables
       */
      MCTreeMatch( const std::string& decay, const std::string& location );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param location1 TES-location of relation tables
       *  @param location2 TES-location of relation tables
       */
      MCTreeMatch( const std::string& decay, const std::string& location1, const std::string& location2 );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param location1 TES-location of relation tables
       *  @param location2 TES-location of relation tables
       *  @param location3 TES-location of relation tables
       */
      MCTreeMatch( const std::string& decay, const std::string& location1, const std::string& location2,
                   const std::string& location3 );
      /** constructor from decay descriptor and TES-locations of tables
       *  @param decay the decay desriptor
       *  @param protoMatch the key for ProtoParticles match
       */
      MCTreeMatch( const std::string& decay, const ProtoPMatch& protoMatch = All, const std::string& thehead = "" );
      // ======================================================================
      /// MANDATORY: virtual destructor
      virtual ~MCTreeMatch();
      /// MANDATORY: clone method ("virtual constructor")
      MCTreeMatch* clone() const override;
      ///
      /// MANDATORY: the only one essential method
      result_type operator()( argument p ) const override;
      /// OPTIONAL: nice printout
      std::ostream& fillStream( std::ostream& ) const override;
      // ======================================================================
    protected:
      // ======================================================================
      /// get MC-particles
      StatusCode getMCParticles( const LHCb::MCParticle::Container& cnt ) const override;
      // ======================================================================
      /// check the validness of finder
      void checkFinder();
      /// get the decay finder
      void getFinder( const std::string& decay );
      // ======================================================================
    private:
      // ======================================================================
      /// default constructor is disabled
      MCTreeMatch(); // default constructor is disabled
      /// the assignement operator is disabled
      MCTreeMatch& operator=( const MCTreeMatch& );
      // ======================================================================
    private:
      // ======================================================================
      /// the actual decay finder
      Finder m_finder; // the actual decay finder
      // ======================================================================
    };
    // ========================================================================
    class MCNodeMatch : public MCMatcherBase {
    public:
      // ======================================================================
      // the actual type of decay node
      typedef Decays::iNode iNode;
      // the actual type of decay node
      typedef Decays::Node Node;
      // ======================================================================
    public:
      // ======================================================================
      /// Default copy constructor
      MCNodeMatch( const MCNodeMatch& ) = default;
      /// constructor from node descriptor and TES-locations of tables
      MCNodeMatch( const iNode& decay, const Locations& locations, const std::string& thehead = "" );
      /// constructor from node descriptor and TES-locations of tables
      MCNodeMatch( const iNode& node, const std::string& location );
      /// constructor from node descriptor and TES-locations of tables
      MCNodeMatch( const iNode& decay, const std::string& location1, const std::string& location2 );
      /// constructor from node descriptor and TES-locations of tables
      MCNodeMatch( const iNode& node, const std::string& location1, const std::string& location2,
                   const std::string& location3 );
      /// constructor from decay descriptor and TES-locations of tables
      MCNodeMatch( const iNode& node, const ProtoPMatch& protoMatch = All, const std::string& thehead = "" );
      // ======================================================================
      /// MANDATORY: virtual destructor
      virtual ~MCNodeMatch();
      /// MANDATORY: clone method ("virtual constructor")
      MCNodeMatch* clone() const override;
      ///
      /// MANDATORY: the only one essential method
      result_type operator()( argument p ) const override;
      /// OPTIONAL: nice printout
      std::ostream& fillStream( std::ostream& ) const override;
      // ======================================================================
    protected:
      // ======================================================================
      /// get MC-particles
      StatusCode getMCParticles( const LHCb::MCParticle::Container& cnt ) const override;
      // ======================================================================
      /// check the validness of node
      void checkNode();
      // ======================================================================
    private:
      // ======================================================================
      /// default constructor is disabled
      MCNodeMatch(); // default constructor is disabled
      /// the assignement operator is disabled
      MCNodeMatch& operator=( const MCNodeMatch& );
      // ======================================================================
    private:
      // ======================================================================
      /// the actual decay node
      Node m_node; // the actual decay node
      // ======================================================================
    };
    // ========================================================================
  } // namespace PhysMCParticles
  // ==========================================================================
  namespace Particles {
    // ========================================================================
    /// import the symbol into proper namespace
    typedef LoKi::PhysMCParticles::MCSelMatch  MCSelMatch;
    typedef LoKi::PhysMCParticles::MCTreeMatch MCTreeMatch;
    typedef LoKi::PhysMCParticles::MCNodeMatch MCNodeMatch;
    // ========================================================================
  } // namespace Particles
  // ==========================================================================
  namespace Cuts {
    // ========================================================================
    /** @typedef MCSELMATCH
     *  helper symbol (for CombineParticles/FilterDesktop framework)
     *  to perform MC-truth match
     *
     *  @code
     *
     *    Cut cut = MCSELMATCH ( "pi+" == MCABSID ) ;
     *
     *  @endcode
     *
     *  @see CombineParticles
     *  @see FilterDesktop
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
     *  @date 2009-08-12
     */
    typedef LoKi::PhysMCParticles::MCSelMatch MCSELMATCH;
    // ========================================================================
    /** @typedef MCTREEMATCH
     *  helper symbol (for CombineParticles/FilterDesktop framework)
     *  to perform MC-truth match
     *
     *  @code
     *
     *    Cut cut = MCTREEMATCH ( "B0 -> pi+ pi-") ;
     *
     *  @endcode
     *
     *  @see CombineParticles
     *  @see FilterDesktop
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
     *  @date 2009-08-12
     */
    typedef LoKi::PhysMCParticles::MCTreeMatch MCTREEMATCH;
    // ========================================================================
    /** @typedef MCNODEMATCH
     *  helper symbol (for CombineParticles/FilterDesktop framework)
     *  to perform MC-truth match
     *
     *  @code
     *
     *    Cut cut = MCNODEMATCH ( StableCharged ) ;
     *
     *  @endcode
     *
     *  @see CombineParticles
     *  @see FilterDesktop
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
     *  @date 2009-08-12
     */
    typedef LoKi::PhysMCParticles::MCNodeMatch MCNODEMATCH;
    // ========================================================================
    /** @typedef MCMATCH
     *  helper symbol (for CombineParticles/FilterDesktop framework)
     *  to perform MC-truth match
     *
     *  @code
     *
     *    Cut cut = MCMATCH ( "B0 -> pi+ pi-") ;
     *
     *  @endcode
     *
     *  @see CombineParticles
     *  @see FilterDesktop
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
     *  @date 2009-08-12
     */
    typedef LoKi::PhysMCParticles::MCTreeMatch MCMATCH;
    // ========================================================================
  } // namespace Cuts
  // ==========================================================================
} //                                                      end of namespace LoKi
// ============================================================================
// The END
// ============================================================================
#endif // LOKI_MCMATCHER_H
